from typing import (
    Union,
    Generator,
    List,
    Optional,
    Callable,
    Any
)
import os
from xopen import xopen
import orjson
from collections.abc import Iterable


def load(
    path: Union[str, bytes, os.PathLike],
    decompression_threads: Optional[int] = None,
    compression_format: Optional[str] = None
) -> List[Union[dict, list, int, float, str, bool, None]]:
    """Deserialize a compressed or uncompressed UTF-8-encoded jsonl file to a list of Python objects.

    Args:
        path (str | bytes | os.PathLike): A path-like object giving the pathname (absolute or relative to the current working directory) of the compressed or uncompressed UTF-8-encoded jsonl file to be deserialized.
        decompression_threads (int, optional): An optional integer passed to xopen.xopen() as the 'threads' argument that specifies the number of threads that should be used for decompression. Defaults to None.
        compression_format (str, optional): An optional string passed to xopen.xopen() as the 'format' argument that overrides the autodetection of the file's compression format based on its extension or content. Possible values are 'gz', 'xz', 'bz2' and 'zst'. Defaults to None.

    Returns:
        list[dict | list | int | float | str | bool | None]: A list of deserialized objects."""

    with xopen(path, 'rb', threads=decompression_threads, format=compression_format) as file:
        return [orjson.loads(json) for json in file]


def stream(
    path: Union[str, bytes, os.PathLike],
    decompression_threads: Optional[int] = None,
    compression_format: Optional[str] = None
) -> Generator[Union[dict, list, int, float, str, bool, None], None, None]:
    """Create a generator that deserializes a compressed or uncompressed UTF-8-encoded jsonl file to Python objects.

    Args:
        path (str | bytes | os.PathLike): A path-like object giving the pathname (absolute or relative to the current working directory) of the compressed or uncompressed UTF-8-encoded jsonl file to be deserialized by the generator.
        decompression_threads (int, optional): An optional integer passed to xopen.xopen() as the 'threads' argument that specifies the number of threads that should be used for decompression. Defaults to None.
        compression_format (str, optional): An optional string passed to xopen.xopen() as the 'format' argument that overrides the autodetection of the file's compression format based on its extension or content. Possible values are 'gz', 'xz', 'bz2' and 'zst'. Defaults to None.

    Returns:
        Generator[dict | list | int | float | str | bool | None, None, None]: A generator that deserializes the file to Python objects."""

    with xopen(path, 'rb', threads=decompression_threads, format=compression_format) as file:
        for json in file:
            yield orjson.loads(json)


def save(
    path: Union[str, bytes, os.PathLike],
    data: Iterable,
    default: Optional[Callable] = None,
    option: int = 0,
    compression_level: Optional[int] = None,
    compression_threads: Optional[int] = None,
    compression_format: Optional[str] = None
) -> None:
    """Serialize an iterable of Python objects to a compressed or uncompressed UTF-8-encoded jsonl file.

    Args:
        path (str | bytes | os.PathLike): A path-like object giving the pathname (absolute or relative to the current working directory) of the compressed or uncompressed UTF-8-encoded jsonl file to be saved.
        data (Iterable): An iterable of Python objects to be serialized to the file.
        default (Callable, optional): An optional callable passed to orjson.dumps() as the 'default' argument that serializes subclasses or arbitrary types to supported types. Defaults to None.
        option (int, optional): An optional integer passed to orjson.dumps() as the 'option' argument that modifies how data is serialized. Defaults to 0.
        compression_level (int, optional): An optional integer passed to xopen.xopen() as the 'compresslevel' argument that determines the compression level for writing to gzip, xz and Zstandard files. Defaults to None.
        compression_threads (int, optional): An optional integer passed to xopen.xopen() as the 'threads' argument that specifies the number of threads that should be used for compression. Defaults to None.
        compression_format (str, optional): An optional string passed to xopen.xopen() as the 'format' argument that overrides the autodetection of the file's compression format based on its extension. Possible values are 'gz', 'xz', 'bz2' and 'zst'. Defaults to None."""

    with xopen(path, 'wb', compresslevel=compression_level, threads=compression_threads, format=compression_format) as writer:
        for item in data:
            writer.write(orjson.dumps(item, default=default, option=option))
            writer.write(b'\n')


def append(
    path: Union[str, bytes, os.PathLike],
    data: Any,
    newline: Optional[bool] = True,
    default: Optional[Callable] = None,
    option: int = 0,
    compression_level: Optional[int] = None,
    compression_threads: Optional[int] = None,
    compression_format: Optional[str] = None
) -> None:
    """Serialize and append a Python object to a compressed or uncompressed UTF-8-encoded jsonl file.

    Args:
        path (str | bytes | os.PathLike): A path-like object giving the pathname (absolute or relative to the current working directory) of the compressed or uncompressed UTF-8-encoded jsonl file to be appended.
        data (Iterable): A Python object to be serialized and appended to the file.
        newline (bool, optional): An optional Boolean flag that, if set to False, indicates that the file does not end with a newline and should, therefore, have one added before data is appended. Defaults to True.
        default (Callable, optional): An optional callable passed to orjson.dumps() as the 'default' argument that serializes subclasses or arbitrary types to supported types. Defaults to None.
        option (int, optional): An optional integer passed to orjson.dumps() as the 'option' argument that modifies how data is serialized. Defaults to 0.
        compression_level (int, optional): An optional integer passed to xopen.xopen() as the 'compresslevel' argument that determines the compression level for writing to gzip, xz and Zstandard files. Defaults to None.
        compression_threads (int, optional): An optional integer passed to xopen.xopen() as the 'threads' argument that specifies the number of threads that should be used for compression. Defaults to None.
        compression_format (str, optional): An optional string passed to xopen.xopen() as the 'format' argument that overrides the autodetection of the file's compression format based on its extension or content. Possible values are 'gz', 'xz', 'bz2' and 'zst'. Defaults to None."""

    with xopen(path, 'ab', compresslevel=compression_level, threads=compression_threads, format=compression_format) as writer:
        if not newline:
            writer.write(b'\n')

        writer.write(orjson.dumps(data, default=default, option=option))
        writer.write(b'\n')


def extend(
    path: Union[str, bytes, os.PathLike],
    data: Iterable,
    newline: Optional[bool] = True,
    default: Optional[Callable] = None,
    option: int = 0,
    compression_level: Optional[int] = None,
    compression_threads: Optional[int] = None,
    compression_format: Optional[str] = None
) -> None:
    """Serialize and append an iterable of Python objects to a compressed or uncompressed UTF-8-encoded jsonl file.

    Args:
        path (str | bytes | os.PathLike): A path-like object giving the pathname (absolute or relative to the current working directory) of the compressed or uncompressed UTF-8-encoded jsonl file to be extended.
        data (Iterable): An iterable of Python objects to be serialized and appended to the file.
        newline (bool, optional): An optional Boolean flag that, if set to False, indicates that the file does not end with a newline and should, therefore, have one added before data is extended. Defaults to True.
        default (Callable, optional): An optional callable passed to orjson.dumps() as the 'default' argument that serializes subclasses or arbitrary types to supported types. Defaults to None.
        option (int, optional): An optional integer passed to orjson.dumps() as the 'option' argument that modifies how data is serialized. Defaults to 0.
        compression_level (int, optional): An optional integer passed to xopen.xopen() as the 'compresslevel' argument that determines the compression level for writing to gzip, xz and Zstandard files. Defaults to None.
        compression_threads (int, optional): An optional integer passed to xopen.xopen() as the 'threads' argument that specifies the number of threads that should be used for compression. Defaults to None.
        compression_format (str, optional): An optional string passed to xopen.xopen() as the 'format' argument that overrides the autodetection of the file's compression format based on its extension or content. Possible values are 'gz', 'xz', 'bz2' and 'zst'. Defaults to None."""

    with xopen(path, 'ab', compresslevel=compression_level, threads=compression_threads, format=compression_format) as writer:
        if not newline:
            writer.write(b'\n')

        for item in data:
            writer.write(orjson.dumps(item, default=default, option=option))
            writer.write(b'\n')



"""Test orjsonl."""

import orjsonl
import helpers
import orjson

DATA = [
    'hello world',
    ['fizz', 'buzz'],
]

JSONL = """\
"hello world"
["fizz","buzz"]
"""


def test_save(tmp_path):
    """Test `orjsonl.save()`."""

    orjsonl.save(tmp_path / 'test_orjsonl.jsonl', data=DATA)
    assert helpers.load(tmp_path / 'test_orjsonl.jsonl') == JSONL.encode('utf-8')


def test_load(tmp_path) -> None:
    """Test `orjsonl.load()`."""
    
    helpers.save(tmp_path / 'test_orjsonl.jsonl', JSONL.encode('utf-8'))
    assert orjsonl.load(tmp_path / 'test_orjsonl.jsonl') == DATA


def test_stream(tmp_path) -> None:
    """Test `orjsonl.stream()`."""
    
    helpers.save(tmp_path / 'test_orjsonl.jsonl', JSONL.encode('utf-8'))
    assert list(orjsonl.stream(tmp_path / 'test_orjsonl.jsonl')) == DATA


def test_append(tmp_path) -> None:
    """Test `orjsonl.append()`."""
    
    helpers.save(tmp_path / 'test_orjsonl.jsonl', JSONL.encode('utf-8'))
    orjsonl.append(tmp_path / 'test_orjsonl.jsonl', data=DATA[-1])
    assert helpers.load(tmp_path / 'test_orjsonl.jsonl') == JSONL.encode('utf-8') + orjson.dumps(DATA[-1]) + b'\n'
    
    helpers.save(tmp_path / 'test_orjsonl.jsonl', JSONL.encode('utf-8'))
    orjsonl.append(tmp_path / 'test_orjsonl.jsonl', data=DATA[-1], newline=False)
    assert helpers.load(tmp_path / 'test_orjsonl.jsonl') == JSONL.encode('utf-8') + b'\n' + orjson.dumps(DATA[-1]) + b'\n'


def test_extend(tmp_path) -> None:
    """Test `orjsonl.extend()`."""

    helpers.save(tmp_path / 'test_orjsonl.jsonl', JSONL.encode('utf-8'))
    orjsonl.extend(tmp_path / 'test_orjsonl.jsonl', data=DATA[-2:])
    assert helpers.load(tmp_path / 'test_orjsonl.jsonl') == JSONL.encode('utf-8') + orjson.dumps(DATA[-2]) + b'\n' + orjson.dumps(DATA[-1]) + b'\n'


    helpers.save(tmp_path / 'test_orjsonl.jsonl', JSONL.encode('utf-8'))
    orjsonl.extend(tmp_path / 'test_orjsonl.jsonl', data=DATA[-2:], newline=False)
    assert helpers.load(tmp_path / 'test_orjsonl.jsonl') == JSONL.encode('utf-8') + b'\n' + orjson.dumps(DATA[-2]) + b'\n' + orjson.dumps(DATA[-1]) + b'\n'


from typing import Any

def save(path: str, data: Any) -> None:
    """Save a file."""
    
    with open(path, 'wb') as file:
        file.write(data)


def load(path: str) -> Any:
    """Load a file."""
    
    with open(path, 'rb') as file:
        return file.read()